#include "ActiveDiskImpl.h"
#include <cstring>
#include <cmath>
#include <sstream>
#include <cstdlib>

using namespace std;

CActiveDisk::CActiveDisk()
   : m_Frequency(1.0), m_DeltaTemp(0.0), normMassVel(0.0), pressValue(0.0), deltaPres(0.0), tempValue(0.0)
{
}

CActiveDisk::~CActiveDisk()
{
}

void CActiveDisk::SetInitParams(INTEROP_COUNT _numParameters, const FVDLL_PARAMETER* _pParameters)
{
   for (INTEROP_COUNT iParam = 0; iParam < _numParameters; ++iParam)
   {
      const FVDLL_PARAMETER & param = _pParameters[iParam];
      if ((strcmp(param.id, "Frequency") == 0) && (param.type == FVDLL_PARAM_REAL))
      {
         m_Frequency = param.value.real;
      }
      else if ((strcmp(param.id, "FlowPress") == 0) && (param.type == FVDLL_PARAM_FILE))
      {
         if (param.value.pFile->size)
         {
            std::stringstream sstr(reinterpret_cast<char*>(param.value.pFile->data));
            std::string buf;

            sstr >> buf;
            while (!sstr.eof())
            {
               if (sstr >> buf)
               {
                  std::size_t pos = buf.find(";");
                  m_J.push_back(std::atof(buf.substr(0, pos).c_str()));
                  m_Kt.push_back(std::atof(buf.substr(pos + 1).c_str()));
               }
            }

            if (m_J.size() != m_Kt.size())
               throw INTEROP_ERROR_INVALID_PARAMETER;
         }
         else
         {
            throw INTEROP_ERROR_INVALID_PARAMETER;
         }
      }
      else if ((strcmp(param.id, "DeltaTemp") == 0) && (param.type == FVDLL_PARAM_REAL))
      {
         m_DeltaTemp = param.value.real;
      }
      else
      {
         throw INTEROP_ERROR_INVALID_PARAMETER;
      }
   }
}

void CActiveDisk::LoadParameters(INTEROP_COUNT _numParameters, const FVDLL_PARAMETER* _pParameters)
{
   for (INTEROP_COUNT iParam = 0; iParam < _numParameters; ++iParam)
   {
      const FVDLL_PARAMETER & param = _pParameters[iParam];
      if ((strcmp(param.id, "normMassVel") == 0) && (param.type == FVDLL_PARAM_REAL))
      {
         normMassVel = param.value.real;
      }
      else
      {
         throw INTEROP_ERROR_INVALID_PARAMETER;
      }
   }
}

void CActiveDisk::SaveParameters()
{
   CheckRes(fvapiSaveParamReal("normMassVel", normMassVel));
}

// Initializing inlet-outlet
void CActiveDisk::InitDisk(FVAPI_BCONDITION _hBCond1, FVAPI_BCONDITION _hBCond2)
{
   // Let's remember inlet and outlet boundary conditions
   m_BConditions[0] = _hBCond1;
   m_BConditions[1] = _hBCond2;

   // Get subregions from boundary conditions
   CheckRes(fvapiGetBConditionSubregion(m_BConditions[0], &m_Subregions[0]));
   CheckRes(fvapiGetBConditionSubregion(m_BConditions[1], &m_Subregions[1]));

   // Get grids from subregions
   CheckRes(fvapiGetSubregionGrid(m_Subregions[0], &m_Grid[0]));
   CheckRes(fvapiGetSubregionGrid(m_Subregions[1], &m_Grid[1]));

   // Request the variable descriptions
   FVAPI_PROCESS hMotion;
   CheckRes(fvapiGetStdProcess(FVAPI_MOTION_PROCESS, &hMotion));
   CheckRes(fvapiRequestVariable(hMotion, "VEL", NULL, &m_Velocity));
   CheckRes(fvapiRequestVariable(hMotion, "PRES", NULL, &m_Pressure));

   FVAPI_PROCESS hHeat;
   CheckRes(fvapiGetStdProcess(FVAPI_HEAT_TRANSFER_PROCESS, &hHeat));
   CheckRes(fvapiRequestVariable(hHeat, "TEMP", NULL, &m_Temperature));

   CheckRes(fvapiRequestVariable(NULL, FVAPI_VARIABLE_DENSITY, NULL, &m_Density));

   tempValue = m_DeltaTemp;
}

// Before time step calculations
void CActiveDisk::BeforeTimeStep()
{
}

// After time step calculations
void CActiveDisk::AfterTimeStep()
{
   INTEROP_DOUBLE time;
   CheckRes(fvapiGetCurrentTime(&time));
   if (time == 0.0)
   {
      pressValue  = GetDeltaPres(normMassVel, time);
      return;
   }

   FVAPI_BOUNDARY_INTEGRAL pResults;
   INTEROP_DOUBLE momentPt[3] = {0.0, 0.0, 0.0};
   if (m_Pressure)
   {
      CheckRes(fvapiGetBoundaryCharacteristics_p(m_Pressure, 1, m_Subregions[0], m_BConditions[0], momentPt, INTEROP_FALSE, &pResults));
      double r = normMassVel;
      normMassVel = (pResults.massFlowMinus - pResults.massFlowPlus) / pResults.area;
      normMassVel += r;
      normMassVel *= 0.5;

      CheckRes(fvapiGetBoundaryCharacteristics_p(m_Pressure, 1, m_Subregions[1], m_BConditions[1], momentPt, INTEROP_FALSE, &pResults));
      pressValue = -GetDeltaPres(normMassVel, time) + pResults.scalarInt;
   }

   if (m_Temperature)
   {
      CheckRes(fvapiGetBoundaryCharacteristics_p(m_Temperature, 1, m_Subregions[0], m_BConditions[0], momentPt, INTEROP_FALSE, &pResults));
      tempValue = m_DeltaTemp + pResults.scalarInt;
   }
}

double CActiveDisk::GetDeltaPres(double &_normMassVel, double _time)
{
   FVAPI_BOUNDARY_INTEGRAL pResults;
   INTEROP_DOUBLE momentPt[3] = {0.0, 0.0, 0.0};
   CheckRes(fvapiGetBoundaryCharacteristics_p(m_Density, 1, m_Subregions[0], m_BConditions[0], momentPt, INTEROP_FALSE, &pResults));

   double ro  = pResults.scalarInt;
   double D   = sqrt(pResults.area * 4. / 3.1415927);
   double rpm = m_Frequency;
   double om  = rpm / 60.;
   double r   = ro * pow(om, 2.) * pow(D, 4.) / pResults.area;

   vector<double> Va, T;
   for (size_t i = 0; i < m_J.size(); ++i)
   {
      Va.push_back(m_J[i] * om * D);
      T.push_back(m_Kt[i] * r);
   }

   if (_time == 0.0)
   {
      int ind = int(Va.size() / 2);
      _normMassVel = Va[ind] * ro;
      return T[ind];
   }

   double vel = _normMassVel / ro;

   if (vel <= Va.front())
      return T.front();

   if (vel >= Va.back())
      return T.back();

   for (size_t i = 1; i < Va.size(); ++i)
   {
      if (Va[i - 1] <= vel && vel < Va[i])
      {
         // Linear interpolation
         double t = (vel - Va[i - 1]) / (Va[i] - Va[i - 1]);
         return (1.0 - t) * T[i - 1] + t * T[i];
      }
   }

   throw INTEROP_ERROR_GENERIC;
}

void CActiveDisk::GetValueInContext(FVAPI_VARIABLE _hVariable, FVAPI_CONTEXT _hVarContext, INTEROP_DOUBLE* _outValue)
{
   if (_hVariable == m_Velocity)
   {
      *_outValue = normMassVel;
   }
   else if (_hVariable == m_Pressure)
   {
      *_outValue = pressValue;
   }
   else if (_hVariable == m_Temperature)
   {
      *_outValue = tempValue;
   }
   else
   {
      throw INTEROP_ERROR_UNKNOWN_ENTITY;
   }
}
